import platform
import subprocess

from .constants import *
from .test import _test
from typing import Iterable, Union

__all__ = [
    "open_file", "save_file", "directory",
    "info", "warning", "error",
    "yesno", "yesnocancel", "retrycancel", "okcancel",
    "YES", "NO", "CANCEL", "RETRY", "OK"
]

SYSTEM = platform.system()

# Find the best dialog for this platform. Default to tkinter.
# Using import keyword instead of __import__ for extra compatibility.
def get_dialogs():
    if SYSTEM == 'Windows':
        from . import windows_dialogs
        return windows_dialogs
    else:
        def cmd_exists(cmd):
            proc = subprocess.Popen(('which', cmd), stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, shell=False)
            proc.communicate()
            return not proc.returncode
        
        if cmd_exists('zenity'):
            from . import zenity_dialogs
            return zenity_dialogs
    
    try:
        from . import tk_dialogs
        return tk_dialogs
    except ModuleNotFoundError: pass
    
    raise ModuleNotFoundError('No dialog type is supported on this machine. Install tkinter to guarantee dialogs.')



dialogs = get_dialogs()

def open_file(title: str = None, filetypes: Iterable[tuple[str, str]] = [("All Files", "*")], multiple: bool = False) -> Union[str, Iterable[str]]:
    '''Shows a dialog box for selecting one or more files to be opened.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.

        filetypes: A list of tuples specifying which filetypes to show.
            The first string is a readable name of that filetype, and
            the second string is one or more glob (e.g., * or *.txt) expression.

            Each glob in the second string is separated by spaces.

            Each tuple will appear in a dropdown of file types to select from.
            If this argument is not specified, all file types are visible.

        multiple: If True, multiple files may be selected. If False, only one file may be selected.

    Returns:
        If `multiple` is True, an iterable of selected files or an empty iterable.
        If `multiple` is False, the file that was selected or an empty string.
    '''
    return dialogs.open_file(title, filetypes, multiple)

def save_file(title: str = None, filetypes: Iterable[tuple[str, str]] = [("All Files", "*")]) -> str:
    '''Shows a dialog box for selecting one or more files to be opened.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.

        filetypes: A list of tuples specifying which filetypes to show.
            The first string is a readable name of that filetype, and
            the second string is one or more glob (e.g., * or *.txt) expression.

            Each glob in the second string is separated by spaces.

            Each tuple will appear in a dropdown of file types to select from.
            If this argument is not specified, all file types are visible.

    Returns:
        The file that was selected or an empty string.
    '''
    return dialogs.save_file(title, filetypes)

def directory(title: str = None) -> str:
    '''Shows a dialog box for selecting a directory. The directory must exist to be selected.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.

    Returns:
        The directory that was selected or an empty string.
    '''
    return dialogs.directory(title)

def info(title: str = None, message: str = '') -> None:
    '''Shows an info dialog box.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.

        message: Text to show in the middle of the dialog box.
    '''
    dialogs.info(title, message)

def warning(title: str = None, message: str = '') -> None:
    '''Shows a warning dialog box.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    '''
    dialogs.warning(title, message)

def error(title: str = None, message: str = '') -> None:
    '''Shows an error dialog box.

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    '''
    dialogs.error(title, message)

def yesno(title: str = None, message: str = '') -> int:
    '''Shows a question dialog box with the buttons "Yes" and "No".

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    
    Returns:
        `xdialog.YES` or `xdialog.NO`. Closing the box results in `xdialog.NO`.
    '''
    return dialogs.yesno(title, message)

def yesnocancel(title: str = None, message: str = '') -> int:
    '''Shows a question dialog box with the buttons "Yes", "No", and "Cancel".

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    
    Returns:
        `xdialog.YES`, `xdialog.NO`, or `xdialog.CANCEL`. Closing the box results in `xdialog.CANCEL`.
    '''
    return dialogs.yesnocancel(title, message)

def retrycancel(title: str = None, message: str = '') -> int:
    '''Shows a question dialog box with the buttons "Retry" and "Cancel".

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    
    Returns:
        `xdialog.RETRY` or `xdialog.CANCEL`. Closing the box results in `xdialog.CANCEL`.
    '''
    return dialogs.retrycancel(title, message)

def okcancel(title: str = None, message: str = '') -> int:
    '''Shows a question dialog box with the buttons "Ok" and "Cancel".

    Arguments:
        title: Text to show on the header of the dialog box.
            Omitting it has system-dependent results.
        
        message: Text to show in the middle of the dialog box.
    
    Returns:
        `xdialog.OK` or `xdialog.CANCEL`. Closing the box results in `xdialog.CANCEL`.
    '''
    return dialogs.okcancel(title, message)


# Code simplified and turned into python bindings from the tk8.6.12/win/tkWinDialog.c file.
# tk is licensed here: https://www.tcl.tk/software/tcltk/license.html

import ctypes 
from ctypes.wintypes import *

LPLPWSTR = ctypes.POINTER(LPWSTR)
LPVOIDP = ctypes.POINTER(ctypes.c_void_p)

LPOFNHOOKPROC = ctypes.WINFUNCTYPE(ctypes.POINTER(UINT), HWND, UINT, WPARAM, LPARAM)


class tagOFNW(ctypes.Structure):
    _fields_ = [
        ("lStructSize",         DWORD),
        ("hwndOwner",           HWND),
        ("hInstance",           HINSTANCE),
        ("lpstrFilter",         LPCWSTR),
        ("lpstrCustomFilter",   LPWSTR),
        ("nMaxCustFilter",      DWORD),
        ("nFilterIndex",        DWORD),
        ("lpstrFile",           LPWSTR),
        ("nMaxFile",            DWORD),
        ("lpstrFileTitle",      LPWSTR),
        ("nMaxFileTitle",       DWORD),
        ("lpstrInitialDir",     LPCWSTR),
        ("lpstrTitle",          LPCWSTR),
        ("Flags",               DWORD),
        ("nFileOffset",         WORD),
        ("nFileExtension",      WORD),
        ("lpstrDefExt",         LPCWSTR),
        ("lCustData",           LPARAM),
        ("lpfnHook",            LPOFNHOOKPROC),
        ("lpTemplateName",      LPCWSTR),
        ("pvReserved",          ctypes.c_void_p),
        ("dwReserved",          DWORD),
        ("FlagsEx",             DWORD)
    ]

    def __init__(self, *args, **kwargs):
        ctypes.memset(ctypes.pointer(self), 0, ctypes.sizeof(tagOFNW))
        super().__init__(*args, **kwargs)


class GUID(ctypes.Structure):
    _fields_ = [
        ("Data1", ctypes.c_ulong),
        ("Data2", ctypes.c_ushort),
        ("Data3", ctypes.c_ushort),
        ("Data4", ctypes.c_ubyte * 8)
    ]

    def __init__(self, data1: int, data2: int, data3: int, data4: tuple[int], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.Data1 = data1
        self.Data2 = data2
        self.Data3 = data3

        arr = self.Data4
        for i in range(8):
            arr[i] = data4[i]

CLSID = GUID
IID = GUID
REFGUID = ctypes.POINTER(GUID)
REFIID = ctypes.POINTER(GUID)

ClsidFileOpenDialog = CLSID(
    0xDC1C5A9C, 0xE88A, 0X4DDE, (0xA5, 0xA1, 0x60, 0xF8, 0x2A, 0x20, 0xAE, 0xF7)
)
IIDIFileOpenDialog = IID(
    0xD57C7288, 0xD4AD, 0x4768, (0xBE, 0x02, 0x9D, 0x96, 0x95, 0x32, 0xD9, 0x60)
)



SFGAOF = ULONG
SICHINTF = DWORD
LPSFGAOF = ctypes.POINTER(SFGAOF)
LPLPOLESTR = ctypes.POINTER(LPOLESTR)
FILEOPENDIALOGOPTIONS = DWORD
LPFILEOPENDIALOGOPTIONS = ctypes.POINTER(FILEOPENDIALOGOPTIONS)

# System dependent enums, could be wrong (I believe this is c_uint)
FDAP = ctypes.c_uint
SIGDN = ctypes.c_uint
SIATTRIBFLAGS = ctypes.c_uint

# Not useful for now
IBindCtx = ctypes.c_void_p
LPIBindCtx = LPVOIDP



# Single shell item class
class IShellItem(ctypes.Structure): pass

LPIShellItem = ctypes.POINTER(IShellItem)
LPLPIShellItem = ctypes.POINTER(ctypes.POINTER(IShellItem))

class IShellItemVtbl(ctypes.Structure):
    _fields_ = [
        ('QueryInterface',  ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, REFIID, LPVOIDP)),
        ('AddRef',          ctypes.WINFUNCTYPE(ULONG, LPIShellItem)),
        ('Release',         ctypes.WINFUNCTYPE(ULONG, LPIShellItem)),
        ('BindToHandler',   ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, LPIBindCtx, REFGUID, REFIID, LPVOIDP)),
        ('GetParent',       ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, LPLPIShellItem)),
        ('GetDisplayName',  ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, SIGDN, LPLPOLESTR)),
        ('GetAttributes',   ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, SFGAOF, LPSFGAOF)),
        ('Compare',         ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItem, LPIShellItem, SICHINTF, ctypes.POINTER(ctypes.c_int))),
    ]

IShellItem.\
    _fields_ = [("lpVtbl", ctypes.POINTER(IShellItemVtbl))]


# Multiple shell item class
class IShellItemArray(ctypes.Structure): pass

LPIShellItemArray = ctypes.POINTER(IShellItem)
LPLPIShellItemArray = ctypes.POINTER(ctypes.POINTER(IShellItem))

class IShellItemArrayVtbl(ctypes.Structure):
    _fields_ = [
        ('QueryInterface',              ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, REFIID, LPVOIDP)),
        ('AddRef',                      ctypes.WINFUNCTYPE(ULONG, LPIShellItemArray)),
        ('Release',                     ctypes.WINFUNCTYPE(ULONG, LPIShellItemArray)),
        ('BindToHandler',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, LPIBindCtx, REFGUID, REFIID, LPVOIDP)),
        ('GetPropertyStore',            ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, ctypes.c_int,  REFIID, LPVOIDP)),
        ('GetPropertyDescriptionList',  ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, ctypes.c_void_p, REFIID, LPVOIDP)),
        ('GetAttributes',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, SIATTRIBFLAGS, SFGAOF, LPSFGAOF)),
        ('GetCount',                    ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, LPDWORD)),
        ('GetItemAt',                   ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, DWORD, LPLPIShellItem)),
        ('EnumItems',                   ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIShellItemArray, LPVOIDP)),
    ]

IShellItemArray.\
    _fields_ = [("lpVtbl", ctypes.POINTER(IShellItemArrayVtbl))]


# file open dialog
class IFileOpenDialog(ctypes.Structure): pass

LPIFileOpenDialog = ctypes.POINTER(IFileOpenDialog)
LPTCLCOMDLG_FILTERSPEC = ctypes.c_void_p

class IFileOpenDialogVtbl(ctypes.Structure):
    _fields_ = [
        ('QueryInterface',          ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, REFIID, LPVOIDP)),
        ('AddRef',                  ctypes.WINFUNCTYPE(ULONG, LPIFileOpenDialog)),
        ('Release',                 ctypes.WINFUNCTYPE(ULONG, LPIFileOpenDialog)),
        ('Show',                    ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, HWND)),
        ('SetFileTypes',            ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, UINT, LPTCLCOMDLG_FILTERSPEC)),
        ('SetFileTypeIndex',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, UINT)),
        ('GetFileTypeIndex',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPUINT)),
        ('Advise',                  ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, ctypes.c_void_p, LPDWORD)),
        ('Unadvise',                ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, DWORD)),
        ('SetOptions',              ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, FILEOPENDIALOGOPTIONS)),
        ('GetOptions',              ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPFILEOPENDIALOGOPTIONS)),
        ('SetDefaultFolder',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPIShellItem)),
        ('SetFolder',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPIShellItem)),
        ('GetFolder',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPIShellItem)),
        ('GetCurrentSelection',     ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPIShellItem)),
        ('SetFileName',             ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPCWSTR)),
        ('GetFileName',             ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPWSTR)),
        ('SetTitle',                ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPCWSTR)),
        ('SetOkButtonLabel',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPCWSTR)),
        ('SetFileNameLabel',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPCWSTR)),
        ('GetResult',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPIShellItem)),
        ('AddPlace',                ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPIShellItem, FDAP)),
        ('SetDefaultExtension',     ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPCWSTR)),
        ('Close',                   ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, ctypes.HRESULT)),
        ('SetClientGuid',           ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, REFGUID)),
        ('ClearClientData',         ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog)),
        ('SetFilter',               ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, ctypes.c_void_p)),
        ('GetResults',              ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPIShellItemArray)),
        ('GetSelectedItems',        ctypes.WINFUNCTYPE(ctypes.HRESULT, LPIFileOpenDialog, LPLPIShellItemArray)),
    ]

IFileOpenDialog.\
    _fields_ = [("lpVtbl", ctypes.POINTER(IFileOpenDialogVtbl))]


from .constants import *

from tkinter import messagebox
from tkinter import filedialog

def open_file(title, filetypes, multiple=False):
    if multiple:
        return filedialog.askopenfilenames(title=title, filetypes=filetypes)
    else:
        return filedialog.askopenfilename(title=title) or ''

def save_file(title, filetypes):
    return filedialog.asksaveasfilename(title=title, filetypes=filetypes) or ''

def directory(title):
    return filedialog.askdirectory(mustexist=True) or ''

info = messagebox.showinfo
warning = messagebox.showwarning
error = messagebox.showerror

def yesno(title, message):
    if messagebox.askyesno(title, message):
        return YES
    else:
        return NO

def yesnocancel(title, message):
    r = messagebox.askyesnocancel(title, message)
    if r is None:
        return CANCEL
    elif r:
        return YES
    else:
        return NO

def retrycancel(title, message):
    if messagebox.askretrycancel(title, message):
        return RETRY
    else:
        return CANCEL

def okcancel(title, message):
    if messagebox.askokcancel(title, message):
        return OK
    else:
        return CANCEL



#!/usr/bin/env python3

import xdialog

from typing import Iterable

def _test():
    print("Beginning test!")

    # Opening Files
    print("Close this.")
    assert xdialog.open_file("Close This!", [("Text File", "*.txt")]) == ''
    print("Select a file.")
    assert len(xdialog.open_file("Select a file", [("Text File", "*.txt"), ("Python File", "*.py")])) > 0
    print("Select multiple files.")
    assert isinstance(xdialog.open_file("Select multiple files.", multiple=True), Iterable)

    # Saving Files
    print("Close this.")
    assert xdialog.save_file("Close This!") == ''
    print("Save to a file.")
    assert len(xdialog.save_file("Select a file to save to", [("Text File", "*.txt")])) > 0

    # Opening a directory
    print("Close this.")
    assert xdialog.directory("Close This!") == ''
    print("Select a directory.")
    assert len(xdialog.directory("Select a directory")) > 0

    # Messages
    xdialog.info("INFO", "Close this box. It should be an info box.")
    xdialog.warning("WARNING", "Close this box. It should be a warning box.")
    xdialog.error("ERROR", "Close this box. It should be an error box.")

    # Questions
    assert xdialog.yesno("YesNo", "Close This! (If you can't, click No)") == xdialog.NO
    assert xdialog.yesno("YesNo", "Click Yes") == xdialog.YES
    assert xdialog.yesno("YesNo", "Click No") == xdialog.NO

    assert xdialog.yesnocancel("YesNoCancel", "Close This! (If you can't, click Cancel)") == xdialog.CANCEL
    assert xdialog.yesnocancel("YesNoCancel", "Click Yes") == xdialog.YES
    assert xdialog.yesnocancel("YesNoCancel", "Click No") == xdialog.NO
    assert xdialog.yesnocancel("YesNoCancel", "Click Cancel") == xdialog.CANCEL

    assert xdialog.retrycancel("RetryCancel", "Close This! (If you can't, click Cancel)") == xdialog.CANCEL
    assert xdialog.retrycancel("RetryCancel", "Click Retry") == xdialog.RETRY
    assert xdialog.retrycancel("RetryCancel", "Click Cancel") == xdialog.CANCEL

    assert xdialog.okcancel("OkCancel", "Close This! (If you can't, click Cancel)") == xdialog.CANCEL
    assert xdialog.okcancel("OkCancel", "Click Ok") == xdialog.OK
    assert xdialog.okcancel("OkCancel", "Click Cancel") == xdialog.CANCEL

    print("Test successful!")


import ctypes
import atexit

from .constants import *
from .windows_structs import *

# dlls
user32 = ctypes.windll.user32
comdlg32 = ctypes.windll.comdlg32
shell32 = ctypes.windll.shell32
ole32 = ctypes.oledll.ole32

BUFFER_SIZE = 8192

# Initialization and uninitialization
ole32.OleInitialize(None)
def oleuninit():
    ole32.OleUninitialize()
atexit.register(oleuninit)


def split_null_list(strp):
    p = ctypes.cast(strp, ctypes.c_wchar_p)
    v = p.value
    while v:
        yield v
        loc = ctypes.cast(p, ctypes.c_void_p).value + (len(v)*2+2)
        p = ctypes.cast(loc, ctypes.c_wchar_p)
        v = p.value


def open_file(title, filetypes, multiple=False):
    file = ctypes.create_unicode_buffer(BUFFER_SIZE)
    pfile = ctypes.cast(file, ctypes.c_wchar_p)

    # Default options
    opts = tagOFNW(
        lStructSize=ctypes.sizeof(tagOFNW),

        lpstrFile=pfile,
        nMaxFile=BUFFER_SIZE,

        lpstrTitle=title,
        Flags=0x00081800 + (0x200 if multiple else 0)
    )

    # Filetypes
    if filetypes:
        out = []
        for s, t in filetypes:
            out.append(f'{s} ({t})\0{";".join(t.split())}\0')
        
        opts.lpstrFilter = ''.join(out)+'\0' # Extra NULL byte just in case
    
    # Call file dialog
    ok = comdlg32.GetOpenFileNameW(ctypes.byref(opts))

    # Return data
    if multiple:
        if ok:
            # Windows splits the parent folder, followed by files, by null characters.
            gen = split_null_list(pfile)
            parent = next(gen)
            return tuple(parent + "\\" + f for f in gen)
        else:
            return ()
    else:
        if ok:
            return file.value
        else:
            return ''

def save_file(title, filetypes):
    file = ctypes.create_unicode_buffer(BUFFER_SIZE)
    pfile = ctypes.cast(file, ctypes.c_wchar_p)

    # Default options
    opts = tagOFNW(
        lStructSize=ctypes.sizeof(tagOFNW),

        lpstrFile=pfile,
        nMaxFile=BUFFER_SIZE,

        lpstrTitle=title,
        Flags=0x00080002
    )

    # Filetypes
    if filetypes:
        out = []
        for s, t in filetypes:
            out.append(f'{s} ({t})\0{";".join(t.split())}\0')
        
        opts.lpstrFilter = ''.join(out)+'\0' # Extra NULL byte just in case
    
    # Call file dialog
    ok = comdlg32.GetSaveFileNameW(ctypes.byref(opts))

    # Return data
    if ok:
        return file.value
    else:
        return ''

# Code simplified and turned into python bindings from the tk8.6.12/win/tkWinDialog.c file.
# tk is licensed here: https://www.tcl.tk/software/tcltk/license.html
def directory(title):
    # Create dialog
    ifd = ctypes.POINTER(IFileOpenDialog)()
    hr = ole32.CoCreateInstance(
        ctypes.byref(ClsidFileOpenDialog),
        None,
        1,
        ctypes.byref(IIDIFileOpenDialog),
        ctypes.byref(ifd)
    )
    if hr < 0: raise OSError("Failed to create dialog")

    # Set options
    flags = UINT(0)
    hr = ifd.contents.lpVtbl.contents.GetOptions(ifd, ctypes.byref(flags))
    if hr < 0: raise OSError("Failed to get options")

    flags = UINT(flags.value | 0x1020)
    hr = ifd.contents.lpVtbl.contents.SetOptions(ifd, flags)

    # Set title
    if title is not None: ifd.contents.lpVtbl.contents.SetTitle(ifd, title)

    try:
        hr = ifd.contents.lpVtbl.contents.Show(ifd, None)
    except OSError:
        return ''
    
    # Acquire selection result
    resultIf = LPIShellItem()
    try:
        hr = ifd.contents.lpVtbl.contents.GetResult(ifd, ctypes.byref(resultIf))
        if hr < 0: raise OSError("Failed to get result of directory selection")
        wstr = LPWSTR()
        hr = resultIf.contents.lpVtbl.contents.GetDisplayName(resultIf, 0x80058000, ctypes.byref(wstr))
        if hr < 0: raise OSError("Failed to get display name from shell item")
        val = wstr.value
        ole32.CoTaskMemFree(wstr)
        return val
    finally:
        resultIf.contents.lpVtbl.contents.Release(resultIf)


# For where the magic numbers come from, see https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-messageboxw

def info(title, message):
    user32.MessageBoxW(None, message or "", title or "Message", 0x00000040)

def warning(title, message):
    user32.MessageBoxW(None, message or "", title or "Warning", 0x00000030)

def error(title, message):
    user32.MessageBoxW(None, message or "", title or "Error", 0x00000010)

def yesno(title, message):
    if user32.MessageBoxW(None, message or "", title or "", 0x00000024) == 6:
        return YES
    else:
        return NO

def yesnocancel(title, message):
    r = user32.MessageBoxW(None, message or "", title or "", 0x00000023)

    if r == 2:
        return CANCEL
    elif r == 6:
        return YES
    else:
        return NO

def retrycancel(title, message):
    if user32.MessageBoxW(None, message or "", title or "", 0x00000025) == 4:
        return RETRY
    else:
        return CANCEL

def okcancel(title, message):
    if user32.MessageBoxW(None, message or "", title or "", 0x00000021) == 1:
        return OK
    else:
        return CANCEL


#!/usr/bin/env python3

import setuptools
import subprocess

from pathlib import Path

build = Path("build.txt")
def get_build():
    if build.is_file():
        num = int(build.read_text())
        build.write_text(str(num+1))
        return num
    else:
        build.write_text("0")
        return 0

setuptools.setup(
    name='xdialog',
    version=f'1.0.5.{get_build()}',
    author='mathgeniuszach',
    author_email='huntingmanzach@gmail.com',
    description='A cross-platform python wrapper for native dialogs.',
    long_description='A cross-platform python wrapper for native dialogs.',
    url='https://github.com/xMGZx/xdialog',
    packages=['xdialog'],
    license='MIT License',
    classifiers=['Programming Language :: Python :: 3']
)




import subprocess

from .constants import *

def clean(txt: str):
    return txt\
        .replace("\\", "\\\\")\
        .replace("$", "\\$")\
        .replace("!", "\\!")\
        .replace("*", "\\*")\
        .replace("?", "\\?")\
        .replace("&", "&amp;")\
        .replace("|", "&#124;")\
        .replace("<", "&lt;")\
        .replace(">", "&gt;")\

def zenity(typ, filetypes=None, **kwargs) -> tuple[int, str]:
    # Build args based on keywords
    args = ['zenity', '--'+typ]
    for k, v in kwargs.items():
        if v is True:
            args.append(f'--{k.replace("_", "-").strip("-")}')
        elif isinstance(v, str):
            cv = clean(v) if k != "title" else v
            args.append(f'--{k.replace("_", "-").strip("-")}={cv}') 
    
    # Build filetypes specially if specified
    if filetypes:
        for name, globs in filetypes:
            if name:
                globlist = globs.split()
                args.append(f'--file-filter={clean(name.replace("|", ""))} ({clean(", ".join(t for t in globlist))})|{clean(globs)}')
    
    proc = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        shell=False
    )
    stdout, _ = proc.communicate()

    return (proc.returncode, stdout.decode('utf-8').strip())


def open_file(title, filetypes, multiple=False):
    if multiple:
        return zenity('file-selection', title=title, filetypes=filetypes, multiple=True, separator="\n")[1].splitlines()
    else:
        return zenity('file-selection', title=title, filetypes=filetypes)[1]

def save_file(title, filetypes):
    return zenity('file-selection', title=title, filetypes=filetypes, save=True)[1]

def directory(title):
    return zenity("file-selection", title=title, directory=True)[1]

def info(title, message):
    zenity("info", title=title, text=message)

def warning(title, message):
    zenity("warning", title=title, text=message)

def error(title, message):
    zenity("error", title=title, text=message)

def yesno(title, message):
    return zenity("question", title=title, text=message)[0]

def yesnocancel(title, message):
    r = zenity(
        "question",
        title=title,
        text=message,
        extra_button="No",
        cancel_label="Cancel"
    )

    if r[1]:
        return NO
    elif r[0]:
        return CANCEL
    else:
        return YES

def retrycancel(title, message):
    r = zenity(
        "question",
        title=(title or "Retry"),
        text=message,
        ok_label="Retry",
        cancel_label="Cancel"
    )[0]

    if r:
        return CANCEL
    else:
        return RETRY

def okcancel(title, message):
    r = zenity(
        "question",
        title=title,
        text=message,
        ok_label="Ok",
        cancel_label="Cancel"
    )[0]

    if r:
        return CANCEL
    else:
        return OK